# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

test_clone <- function(mat) {
    .Call('_morebeachtests_test_clone', PACKAGE = 'morebeachtests', mat)
}

test_clone_sparse <- function(mat) {
    .Call('_morebeachtests_test_clone_sparse', PACKAGE = 'morebeachtests', mat)
}

test_sparse_writer1 <- function(type) {
    .Call('_morebeachtests_test_sparse_writer1', PACKAGE = 'morebeachtests', type)
}

test_sparse_writer2 <- function(mat, replacement) {
    .Call('_morebeachtests_test_sparse_writer2', PACKAGE = 'morebeachtests', mat, replacement)
}

test_sparse_writer3 <- function() {
    .Call('_morebeachtests_test_sparse_writer3', PACKAGE = 'morebeachtests')
}

get_column_slice <- function(mat, order, starts, ends, mode) {
    .Call('_morebeachtests_get_column_slice', PACKAGE = 'morebeachtests', mat, order, starts, ends, mode)
}

get_column <- function(mat, order, mode) {
    .Call('_morebeachtests_get_column', PACKAGE = 'morebeachtests', mat, order, mode)
}

get_row_slice <- function(mat, order, starts, ends, mode) {
    .Call('_morebeachtests_get_row_slice', PACKAGE = 'morebeachtests', mat, order, starts, ends, mode)
}

get_row <- function(mat, order, mode) {
    .Call('_morebeachtests_get_row', PACKAGE = 'morebeachtests', mat, order, mode)
}

get_sparse_column_slice <- function(mat, order, starts, ends, mode) {
    .Call('_morebeachtests_get_sparse_column_slice', PACKAGE = 'morebeachtests', mat, order, starts, ends, mode)
}

get_sparse_column <- function(mat, order, mode) {
    .Call('_morebeachtests_get_sparse_column', PACKAGE = 'morebeachtests', mat, order, mode)
}

get_sparse_row_slice <- function(mat, order, starts, ends, mode) {
    .Call('_morebeachtests_get_sparse_row_slice', PACKAGE = 'morebeachtests', mat, order, starts, ends, mode)
}

get_sparse_row <- function(mat, order, mode) {
    .Call('_morebeachtests_get_sparse_row', PACKAGE = 'morebeachtests', mat, order, mode)
}

test_promotion <- function(mat) {
    .Call('_morebeachtests_test_promotion', PACKAGE = 'morebeachtests', mat)
}

