#ifndef TATAMI_R_TATAMI_R_HPP
#define TATAMI_R_TATAMI_R_HPP

#include "UnknownMatrix.hpp"
#include "parallelize.hpp"

/** 
 * @file tatami_r.hpp
 * @brief **tatami** bindings for arbitrary R matrices.
 */

/**
 * @namespace tatami_r
 * @brief **tatami** bindings for arbitrary R matrices.
 */
namespace tatami_r {}

#endif
