#ifndef TATAMI_CHUNKED_TATAMI_CHUNKED_HPP
#define TATAMI_CHUNKED_TATAMI_CHUNKED_HPP

/**
 * @file tatami_chunked.hpp
 * @brief Umbrella header for chunked tatami matrices.
 */

#include "CustomChunkedMatrix.hpp"
#include "LruSlabCache.hpp"
#include "OracleSlabCache.hpp"
#include "simple_chunk_wrappers.hpp"
#include "typical_slab_cache.hpp"

/**
 * @namespace tatami_chunked
 * @brief Methods to handle chunked tatami matrices.
 */
namespace tatami_chunked {}

#endif
